@@ -1,10 +1,14 @@
 from __future__ import unicode_literals
 
-from ctypes import windll, byref, ArgumentError, c_char, c_long, c_ulong
+from ctypes import windll, byref, ArgumentError, c_char, c_long, c_ulong, c_uint, pointer
 from ctypes.wintypes import DWORD
-from ..win32_types import CONSOLE_SCREEN_BUFFER_INFO, STD_OUTPUT_HANDLE, COORD, SMALL_RECT
 
-import sys
+from prompt_toolkit.renderer import Output
+from prompt_toolkit.styles import ANSI_COLOR_NAMES
+from prompt_toolkit.win32_types import CONSOLE_SCREEN_BUFFER_INFO, STD_OUTPUT_HANDLE, STD_INPUT_HANDLE, COORD, SMALL_RECT
+
+import os
+import six
 
 __all__ = (
     'Win32Output',
@@ -36,8 +40,30 @@ _DEBUG_RENDER_OUTPUT = False
 _DEBUG_RENDER_OUTPUT_FILENAME = r'prompt-toolkit-windows-output.log'
 
 
-class Win32Output(object):
+class NoConsoleScreenBufferError(Exception):
     """
+    Raised when the application is not running inside a Windows Console, but
+    the user tries to instantiate Win32Output.
+    """
+    def __init__(self):
+        # Are we running in 'xterm' on Windows, like git-bash for instance?
+        xterm = 'xterm' in os.environ.get('TERM')
+
+        if xterm:
+            message = ('Found %s, while expecting a Windows console. '
+                       'Maybe try to run this program using "winpty" '
+                       'or run it in cmd.exe instead. Or otherwise, '
+                       'in case of Cygwin, use the Python executable '
+                       'that is compiled for Cygwin.' % os.environ['TERM'])
+        else:
+            message = 'No Windows console found. Are you running cmd.exe?'
+        super(NoConsoleScreenBufferError, self).__init__(message)
+
+
+class Win32Output(Output):
+    """
+    I/O abstraction for rendering to Windows consoles.
+    (cmd.exe and similar.)
     """
     def __init__(self, stdout, use_complete_width=False):
         self.use_complete_width = use_complete_width
@@ -46,17 +72,35 @@ class Win32Output(object):
         self.stdout = stdout
         self.hconsole = windll.kernel32.GetStdHandle(STD_OUTPUT_HANDLE)
 
+        self._in_alternate_screen = False
+
         self.color_lookup_table = ColorLookupTable()
 
+        # Remember the default console colors.
+        info = self.get_win32_screen_buffer_info()
+        self.default_attrs = info.wAttributes if info else 15
+
         if _DEBUG_RENDER_OUTPUT:
             self.LOG = open(_DEBUG_RENDER_OUTPUT_FILENAME, 'ab')
 
+    def fileno(self):
+        " Return file descriptor. "
+        return self.stdout.fileno()
+
+    def encoding(self):
+        " Return encoding used for stdout. "
+        return self.stdout.encoding
+
     def write(self, data):
         self._buffer.append(data)
 
+    def write_raw(self, data):
+        " For win32, there is no difference between write and write_raw. "
+        self.write(data)
+
     def get_size(self):
-        from ..renderer import Size
-        info = self._screen_buffer_info()
+        from prompt_toolkit.layout.screen import Size
+        info = self.get_win32_screen_buffer_info()
 
         # We take the width of the *visible* region as the size. Not the width
         # of the complete screen buffer. (Unless use_complete_width has been
@@ -66,7 +110,7 @@ class Win32Output(object):
         else:
             width = info.srWindow.Right - info.srWindow.Left
 
-        height = info.srWindow.Bottom - info.srWindow.Top
+        height = info.srWindow.Bottom - info.srWindow.Top + 1
 
         # We avoid the right margin, windows will wrap otherwise.
         maxwidth = info.dwSize.X - 1
@@ -93,26 +137,38 @@ class Win32Output(object):
             if _DEBUG_RENDER_OUTPUT:
                 self.LOG.write(('    Error in %r %r %s\n' % (func.__name__, e, e)).encode('utf-8'))
 
-    def _screen_buffer_info(self):
+    def get_win32_screen_buffer_info(self):
         """
         Return Screen buffer info.
         """
         sbinfo = CONSOLE_SCREEN_BUFFER_INFO()
         success = self._winapi(windll.kernel32.GetConsoleScreenBufferInfo,
-                     self.hconsole, byref(sbinfo))
+                               self.hconsole, byref(sbinfo))
         if success:
             return sbinfo
+        else:
+            raise NoConsoleScreenBufferError
+
+    def set_title(self, title):
+        """
+        Set terminal title.
+        """
+        assert isinstance(title, six.text_type)
+        self._winapi(windll.kernel32.SetConsoleTitleW, title)
+
+    def clear_title(self):
+        self._winapi(windll.kernel32.SetConsoleTitleW, '')
 
     def erase_screen(self):
         start = COORD(0, 0)
-        sbinfo = self._screen_buffer_info()
+        sbinfo = self.get_win32_screen_buffer_info()
         length = sbinfo.dwSize.X * sbinfo.dwSize.Y
 
         self.cursor_goto(row=0, column=0)
         self._erase(start, length)
 
     def erase_down(self):
-        sbinfo = self._screen_buffer_info()
+        sbinfo = self.get_win32_screen_buffer_info()
         size = sbinfo.dwSize
 
         start = sbinfo.dwCursorPosition
@@ -123,7 +179,7 @@ class Win32Output(object):
     def erase_end_of_line(self):
         """
         """
-        sbinfo = self._screen_buffer_info()
+        sbinfo = self.get_win32_screen_buffer_info()
         start = sbinfo.dwCursorPosition
         length = sbinfo.dwSize.X - sbinfo.dwCursorPosition.X
 
@@ -137,15 +193,28 @@ class Win32Output(object):
                      byref(chars_written))
 
         # Reset attributes.
-        sbinfo = self._screen_buffer_info()
+        sbinfo = self.get_win32_screen_buffer_info()
         self._winapi(windll.kernel32.FillConsoleOutputAttribute,
                      self.hconsole, sbinfo.wAttributes, length, _coord_byval(start),
                      byref(chars_written))
 
     def reset_attributes(self):
-        self._winapi(windll.kernel32.SetConsoleTextAttribute, self.hconsole, 15)  # White
+        " Reset the console foreground/background color. "
+        self._winapi(windll.kernel32.SetConsoleTextAttribute, self.hconsole,
+                     self.default_attrs)
+
+    def set_attributes(self, attrs):
+        fgcolor, bgcolor, bold, underline, italic, blink, reverse = attrs
+
+        if reverse:
+            fgcolor, bgcolor = bgcolor, fgcolor
+
+            # Make sure to reverse, even when no values were specified.
+            if fgcolor is None:
+                fgcolor = '000000'
+            if bgcolor is None:
+                bgcolor = 'ffffff'
 
-    def set_attributes(self, fgcolor, bgcolor, bold=False, underline=False):
         i = self.color_lookup_table.lookup_color(fgcolor, bgcolor)
         self._winapi(windll.kernel32.SetConsoleTextAttribute, self.hconsole, i)
 
@@ -159,25 +228,22 @@ class Win32Output(object):
 
     def cursor_goto(self, row=0, column=0):
         pos = COORD(x=column, y=row)
-        self._winapi(windll.kernel32.SetConsoleCursorPosition,
-                     self.hconsole, _coord_byval(pos))
+        self._winapi(windll.kernel32.SetConsoleCursorPosition, self.hconsole, _coord_byval(pos))
 
     def cursor_up(self, amount):
-        sr = self._screen_buffer_info().dwCursorPosition
+        sr = self.get_win32_screen_buffer_info().dwCursorPosition
         pos = COORD(sr.X, sr.Y - amount)
-        success = self._winapi(windll.kernel32.SetConsoleCursorPosition,
-                               self.hconsole, _coord_byval(pos))
+        self._winapi(windll.kernel32.SetConsoleCursorPosition, self.hconsole, _coord_byval(pos))
 
     def cursor_down(self, amount):
         self.cursor_up(-amount)
 
     def cursor_forward(self, amount):
-        sr = self._screen_buffer_info().dwCursorPosition
-        assert sr.X + amount >= 0, 'Negative cursor position: x=%r amount=%r' % (sr.X, amount)
+        sr = self.get_win32_screen_buffer_info().dwCursorPosition
+#        assert sr.X + amount >= 0, 'Negative cursor position: x=%r amount=%r' % (sr.X, amount)
 
         pos = COORD(max(0, sr.X + amount), sr.Y)
-        success = self._winapi(windll.kernel32.SetConsoleCursorPosition,
-                               self.hconsole, _coord_byval(pos))
+        self._winapi(windll.kernel32.SetConsoleCursorPosition, self.hconsole, _coord_byval(pos))
 
     def cursor_backward(self, amount):
         self.cursor_forward(-amount)
@@ -199,15 +265,20 @@ class Win32Output(object):
             self.LOG.write(('%r' % data).encode('utf-8') + b'\n')
             self.LOG.flush()
 
-        # Write to output
-        self.stdout.write(data)
-        self.stdout.flush()
+        # Print characters one by one. This appears to be the best soluton
+        # in oder to avoid traces of vertical lines when the completion
+        # menu disappears.
+        for b in data:
+            written = DWORD()
+
+            retval = windll.kernel32.WriteConsoleW(self.hconsole, b, 1, byref(written), None)
+            assert retval != 0
 
         self._buffer = []
 
     def get_rows_below_cursor_position(self):
-        info = self._screen_buffer_info()
-        return info.srWindow.Bottom - info.dwCursorPosition.Y# - 1
+        info = self.get_win32_screen_buffer_info()
+        return info.srWindow.Bottom - info.dwCursorPosition.Y + 1
 
     def scroll_buffer_to_prompt(self):
         """
@@ -215,7 +286,7 @@ class Win32Output(object):
         to left, with the cursor at the bottom (if possible).
         """
         # Get current window size
-        info = self._screen_buffer_info()
+        info = self.get_win32_screen_buffer_info()
         sr = info.srWindow
         cursor_pos = info.dwCursorPosition
 
@@ -231,8 +302,71 @@ class Win32Output(object):
         result.Top = result.Bottom - win_height
 
         # Scroll API
-        success = self._winapi(windll.kernel32.SetConsoleWindowInfo,
-                               self.hconsole, True, byref(result))
+        self._winapi(windll.kernel32.SetConsoleWindowInfo, self.hconsole, True, byref(result))
+
+    def enter_alternate_screen(self):
+        """
+        Go to alternate screen buffer.
+        """
+        if not self._in_alternate_screen:
+            GENERIC_READ = 0x80000000
+            GENERIC_WRITE = 0x40000000
+
+            # Create a new console buffer and activate that one.
+            handle = self._winapi(windll.kernel32.CreateConsoleScreenBuffer, GENERIC_READ|GENERIC_WRITE,
+                                  DWORD(0), None, DWORD(1), None)
+
+            self._winapi(windll.kernel32.SetConsoleActiveScreenBuffer, handle)
+            self.hconsole = handle
+            self._in_alternate_screen = True
+
+    def quit_alternate_screen(self):
+        """
+        Make stdout again the active buffer.
+        """
+        if self._in_alternate_screen:
+            stdout = self._winapi(windll.kernel32.GetStdHandle, STD_OUTPUT_HANDLE)
+            self._winapi(windll.kernel32.SetConsoleActiveScreenBuffer, stdout)
+            self._winapi(windll.kernel32.CloseHandle, self.hconsole)
+            self.hconsole = stdout
+            self._in_alternate_screen = False
+
+    def enable_mouse_support(self):
+        ENABLE_MOUSE_INPUT = 0x10
+        handle = windll.kernel32.GetStdHandle(STD_INPUT_HANDLE)
+
+        original_mode = DWORD()
+        self._winapi(windll.kernel32.GetConsoleMode, handle, pointer(original_mode))
+        self._winapi(windll.kernel32.SetConsoleMode, handle, original_mode.value | ENABLE_MOUSE_INPUT)
+
+    def disable_mouse_support(self):
+        ENABLE_MOUSE_INPUT = 0x10
+        handle = windll.kernel32.GetStdHandle(STD_INPUT_HANDLE)
+
+        original_mode = DWORD()
+        self._winapi(windll.kernel32.GetConsoleMode, handle, pointer(original_mode))
+        self._winapi(windll.kernel32.SetConsoleMode, handle, original_mode.value & ~ ENABLE_MOUSE_INPUT)
+
+    def hide_cursor(self):
+        pass
+
+    def show_cursor(self):
+        pass
+
+    @classmethod
+    def win32_refresh_window(cls):
+        """
+        Call win32 API to refresh the whole Window.
+
+        This is sometimes necessary when the application paints background
+        for completion menus. When the menu disappears, it leaves traces due
+        to a bug in the Windows Console. Sending a repaint request solves it.
+        """
+        # Get console handle
+        handle = windll.kernel32.GetConsoleWindow()
+
+        RDW_INVALIDATE = 0x0001
+        windll.user32.RedrawWindow(handle, None, None, c_uint(RDW_INVALIDATE))
 
 
 class FOREGROUND_COLOR:
@@ -243,7 +377,7 @@ class FOREGROUND_COLOR:
     RED       = 0x0004
     MAGENTA   = 0x0005
     YELLOW    = 0x0006
-    GREY      = 0x0007
+    GRAY      = 0x0007
     INTENSITY = 0x0008  # Foreground color is intensified.
 
 
@@ -255,10 +389,44 @@ class BACKROUND_COLOR:
     RED       = 0x0040
     MAGENTA   = 0x0050
     YELLOW    = 0x0060
-    GREY      = 0x0070
+    GRAY      = 0x0070
     INTENSITY = 0x0080  # Background color is intensified.
 
 
+def _create_ansi_color_dict(color_cls):
+    " Create a table that maps the 16 named ansi colors to their Windows code. "
+    return {
+        'ansiblack':   color_cls.BLACK,
+        'ansidefault': color_cls.BLACK,
+        'ansiwhite':   color_cls.GRAY | color_cls.INTENSITY,
+
+        # Low intensity.
+        'ansired':         color_cls.RED,
+        'ansigreen':       color_cls.GREEN,
+        'ansiyellow':      color_cls.YELLOW,
+        'ansiblue':        color_cls.BLUE,
+        'ansifuchsia':     color_cls.MAGENTA,
+        'ansiturquoise':   color_cls.CYAN,
+        'ansilightgray':   color_cls.GRAY,
+
+
+        # High intensity.
+        'ansidarkgray':    color_cls.BLACK | color_cls.INTENSITY,
+        'ansidarkred':     color_cls.RED | color_cls.INTENSITY,
+        'ansidarkgreen':   color_cls.GREEN | color_cls.INTENSITY,
+        'ansibrown':       color_cls.YELLOW | color_cls.INTENSITY,
+        'ansidarkblue':    color_cls.BLUE | color_cls.INTENSITY,
+        'ansipurple':      color_cls.MAGENTA | color_cls.INTENSITY,
+        'ansiteal':        color_cls.CYAN | color_cls.INTENSITY,
+    }
+
+FG_ANSI_COLORS = _create_ansi_color_dict(FOREGROUND_COLOR)
+BG_ANSI_COLORS = _create_ansi_color_dict(BACKROUND_COLOR)
+
+assert set(FG_ANSI_COLORS) == set(ANSI_COLOR_NAMES)
+assert set(BG_ANSI_COLORS) == set(ANSI_COLOR_NAMES)
+
+
 class ColorLookupTable(object):
     """
     Inspired by pygments/formatters/terminal256.py
@@ -267,7 +435,8 @@ class ColorLookupTable(object):
         self._win32_colors = self._build_color_table()
         self.best_match = {}  # Cache
 
-    def _build_color_table(self):
+    @staticmethod
+    def _build_color_table():
         """
         Build an RGB-to-256 color conversion table
         """
@@ -282,7 +451,7 @@ class ColorLookupTable(object):
             (0xaa, 0x00, 0x00, FG.RED, BG.RED),
             (0xaa, 0x00, 0xaa, FG.MAGENTA, BG.MAGENTA),
             (0xaa, 0xaa, 0x00, FG.YELLOW, BG.YELLOW),
-            (0x88, 0x88, 0x88, FG.GREY, BG.GREY),
+            (0x88, 0x88, 0x88, FG.GRAY, BG.GRAY),
 
             (0x44, 0x44, 0xff, FG.BLUE | FG.INTENSITY, BG.BLUE | BG.INTENSITY),
             (0x44, 0xff, 0x44, FG.GREEN | FG.INTENSITY, BG.GREEN | BG.INTENSITY),
@@ -291,7 +460,8 @@ class ColorLookupTable(object):
             (0xff, 0x44, 0xff, FG.MAGENTA | FG.INTENSITY, BG.MAGENTA | BG.INTENSITY),
             (0xff, 0xff, 0x44, FG.YELLOW | FG.INTENSITY, BG.YELLOW | BG.INTENSITY),
 
-            (0xff, 0xff, 0xff, FG.GREY | FG.INTENSITY, BG.GREY | BG.INTENSITY),
+            (0x44, 0x44, 0x44, FG.BLACK | FG.INTENSITY, BG.BLACK | BG.INTENSITY),
+            (0xff, 0xff, 0xff, FG.GRAY | FG.INTENSITY, BG.GRAY | BG.INTENSITY),
         ]
 
     def _closest_color(self, r, g, b):
@@ -328,10 +498,28 @@ class ColorLookupTable(object):
         return indexes
 
     def lookup_color(self, fg_color, bg_color):
-        fg_index = self._color_indexes(fg_color)[0]
-        bg_index = self._color_indexes(bg_color)[1]
+        """
+        Return the color for use in the
+        `windll.kernel32.SetConsoleTextAttribute` API call.
 
-        if fg_index == 0:
-            fg_index = FOREGROUND_COLOR.GREY
+        :param fg_color: Foreground as text. E.g. 'ffffff' or 'red'
+        :param bg_color: Background as text. E.g. 'ffffff' or 'red'
+        """
+        # Set the default foreground color. (Otherwise, many things will be
+        # invisible.) Note that gray is the default on Windows, not GRAY|INTENSITY!
+        if fg_color is None:
+            fg_index = FOREGROUND_COLOR.GRAY
+        else:
+            # Foreground.
+            if fg_color in FG_ANSI_COLORS:
+                fg_index = FG_ANSI_COLORS[fg_color]
+            else:
+                fg_index = self._color_indexes(fg_color)[0]
+
+        # Background.
+        if bg_color in BG_ANSI_COLORS:
+            bg_index = BG_ANSI_COLORS[bg_color]
+        else:
+            bg_index = self._color_indexes(bg_color)[1]
 
         return fg_index | bg_index
